%% File: mleancop_ds5_10.pl  -  Version: 1.0  -  Date: 15 May 2022
%%
%%         "Make everything as simple as possible, but not simpler."
%%                                                 [Albert Einstein]
%%
%% Purpose: MleanCoP-DS5: A Connection Prover for Modal Logic
%%
%% Author:  Jens Otten
%% Web:     www.leancop.de/mleancop/
%%
%% Usage:   prove(F,P).      % where F is a modal first-order formula
%%                           %  and P is the returned connection proof
%%                           %  e.g. F=((# all X:p(X)) => all X: #p(X))
%%                           %  and  P=[[p(3^[]^[]):[4^[]]],
%%                           %         [[-(p(3^[]^[])):-([[4^[]]])]]]
%%          prove2(F,S,P).   % where F is a formula, S is a subset of
%%                           %  [nodef,def,conj,reo(I),scut,cut,comp(J)]
%%                           %  (with numbers I,J) defining attributes
%%                           %  and P is the returned connection proof
%%
%% Copyright: (c) 2011-2022 by Jens Otten
%% License:   GNU General Public License

logic(s5).       % specify modal logic (d,s5)
domain(cumul).   % specify domain condition (const,cumul,vary)


:- lib(iso).                   % load library for ISO compatibility
:- set_flag(occur_check,on).   % global occur check on
:- [mleancop_defmm].           % load program for clausal form translation
:- dynamic(pathlim/0), dynamic(lit/3).

%%% prove formula F

prove(F,Proof) :- prove2(F,[scut,cut,comp(7)],Proof).

prove2(F,Set,Proof) :-
    (F=[_|_] -> M=F ; make_matrix_modal(F,Set,M)),
    retract_all(lit(_,_,_)), (member([(-(#)):_],M) -> S=conj ; S=pos),
    assert_clauses(M,S), prove(1,Set,Proof).

prove(PathLim,Set,Proof) :-
    ( \+member(scut,Set) ->
      prove([(-(#)):(-[])],[],PathLim,[],[],_,Set,[Proof]) ;
      lit((#):_,FV:C,_) ->
      prove(C,[(-(#)):(-[])],PathLim,[],FV,_,Set,Proof1),
      Proof=[C|Proof1] ).
prove(PathLim,Set,Proof) :-
    member(comp(Limit),Set), PathLim=Limit -> prove(1,[],Proof) ;
    (member(comp(_),Set);retract(pathlim)) ->
    PathLim1 is PathLim+1, prove(PathLim1,Set,Proof).

%%% MleanCoP core prover

prove([],_,_,_,FV,FV,_,[]).

prove([Lit:Pre|Cla],Path,PathLim,Lem,FreeV,FreeV1,Set,Proof) :-
    Proof=[[[NegLit:PreN|Cla1]|Proof1]|Proof2],
    \+ (member(LitC,[Lit:Pre|Cla]), member(LitP,Path), LitC==LitP),
    (-NegLit=Lit;-Lit=NegLit) ->
       ( member(LitL,Lem), Lit:Pre==LitL, Cla1=[], Proof1=[],
         FreeV3=FreeV
         ;
         member(NegL:PreN,Path), unify_with_occurs_check(NegL,NegLit),
         Cla1=[], Proof1=[],
         prefix_unify([Pre=PreN]), domain_cond(FreeV), FreeV3=FreeV
         ;
         lit(NegLit:PreN,FV:Cla1,Grnd1),
         ( Grnd1=g -> true ; length(Path,K), K<PathLim -> true ;
           \+ pathlim -> assert(pathlim), fail ),
         prefix_unify([Pre=PreN]),
         append(FreeV,FV,FreeV2), domain_cond(FreeV2),
         prove(Cla1,[Lit:Pre|Path],PathLim,Lem,FreeV2,FreeV3,Set,Proof1)
       ),
       ( member(cut,Set) -> ! ; true ),
       prove(Cla,Path,PathLim,[Lit:Pre|Lem],FreeV3,FreeV1,Set,Proof2).

%%% write clauses into Prolog's database

assert_clauses([],_).
assert_clauses([FV:C|M],Set) :-
    (Set\=conj, \+member((-_):_,C) -> C1=[(#):[]|C] ; C1=C),
    (ground(C) -> G=g ; G=n), assert_clauses2(C1,[],G,FV),
    assert_clauses(M,Set).

assert_clauses2([],_,_,_).
assert_clauses2([L|C],C1,G,FV) :-
    append(C1,C,C2), assert(lit(L,FV:C2,G)), append(C1,[L],C3),
    assert_clauses2(C,C3,G,FV).


%%% prefix unification for D, S5

prefix_unify([]).
prefix_unify([S=T|G]) :-
    ( -S2=S -> T2=T ; -S2=T, T2=S ),
    ( logic(s5)    -> S1=S2, T1=T2 ; flatten(S2,S1), flatten(T2,T1) ),
    logic(Logic) -> prefix_unifyL(S1,T1,Logic), prefix_unify(G).

prefix_unifyL(S,T,Logic) :-
    ( Logic=d     -> tuni_d(S,T) ;
      Logic=s5    -> tuni_s5(S,T) ).

%%% rules for D

tuni_d([],[]).
tuni_d([X1|S],[X2|T]) :- unify_with_occurs_check(X1,X2), tuni_d(S,T).

%%% rules for S5

tuni_s5([],[]).
tuni_s5(S,[]) :- append(_,[X1],S), X1=[].
tuni_s5([],T) :- append(_,[X1],T), X1=[].
tuni_s5(S,T)  :- append(_,[X1],S), append(_,[X2],T),
                 unify_with_occurs_check(X1,X2).

%%% check domain condition

domain_cond(FV) :- domain(const) -> true ; domcond(FV,FV).

domcond([],_).
domcond([[X,Pre]|L],FV) :- domco(X,Pre,FV), domcond(L,FV).

domco(X,_,_)    :- (atomic(X);X==[[]]), !.
domco(X,Pre,FV) :- var(X), !, ( \+ domain(vary) -> true ;
                   domcom(X,FV,Pre1), dom_unify(Pre1,Pre) ).
domco(_^_^Pre1,Pre,_) :- !, dom_unify(Pre1,Pre).
domco(T,Pre,FV) :- T=..[_,U|V], domco(U,Pre,FV), domco(V,Pre,FV).

domcom(X,[[X1,Pre1]|FV],Pre) :- X==X1 -> Pre=Pre1 ; domcom(X,FV,Pre).

dom_unify(Pre1,Pre2) :-
    ( logic(s5)    -> S=Pre1, T=Pre2 ; flatten(Pre1,S), flatten(Pre2,T) ),
    logic(Logic) ->
    ( domain(vary)  -> prefix_unifyL(S,T,Logic) ;
      domain(cumul) -> dom_unifyLcu(S,T,Logic) ).

dom_unifyLcu(S,T,Logic) :-
      ( Logic=d     -> length(S,LenS), append(T1,_,T), length(T1,LenS),
                       tuni_d(S,T1) ;
        Logic=s5    -> true ).
